/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.commons.jxpath.ri.model.jdom;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.commons.jxpath.ri.QName;
import org.apache.commons.jxpath.ri.model.NodeIterator;
import org.apache.commons.jxpath.ri.model.NodePointer;
import org.jdom.Attribute;
import org.jdom.Element;
import org.jdom.Namespace;

/**
 * An iterator of attributes of a DOM Node.
 */
public class JDOMAttributeIterator implements NodeIterator {

    private NodePointer parent;
    private List<Attribute> attributes;
    private int position;

    /**
     * Constructs a new JDOMAttributeIterator.
     *
     * @param parent pointer
     * @param qName   test
     */
    public JDOMAttributeIterator(final NodePointer parent, final QName qName) {
        this.parent = parent;
        if (parent.getNode() instanceof Element) {
            final Element element = (Element) parent.getNode();
            final String prefix = qName.getPrefix();
            Namespace ns = null;
            if (prefix != null) {
                if (prefix.equals("xml")) {
                    ns = Namespace.XML_NAMESPACE;
                } else {
                    final String uri = parent.getNamespaceResolver().getNamespaceURI(prefix);
                    if (uri != null) {
                        ns = Namespace.getNamespace(prefix, uri);
                    }
                    if (ns == null) {
                        // TBD: no attributes
                        attributes = Collections.emptyList();
                        return;
                    }
                }
            } else {
                ns = Namespace.NO_NAMESPACE;
            }
            final String lname = qName.getName();
            if (!lname.equals("*")) {
                attributes = new ArrayList<>();
                final Attribute attr = element.getAttribute(lname, ns);
                if (attr != null) {
                    attributes.add(attr);
                }
            } else {
                attributes = new ArrayList<>();
                final List<Attribute> allAttributes = element.getAttributes();
                for (int i = 0; i < allAttributes.size(); i++) {
                    final Attribute attr = allAttributes.get(i);
                    if (ns == Namespace.NO_NAMESPACE || attr.getNamespace().equals(ns)) {
                        attributes.add(attr);
                    }
                }
            }
        }
    }

    @Override
    public NodePointer getNodePointer() {
        if (position == 0) {
            if (!setPosition(1)) {
                return null;
            }
            position = 0;
        }
        int index = position - 1;
        if (index < 0) {
            index = 0;
        }
        return new JDOMAttributePointer(parent, attributes.get(index));
    }

    @Override
    public int getPosition() {
        return position;
    }

    @Override
    public boolean setPosition(final int position) {
        if (attributes == null) {
            return false;
        }
        this.position = position;
        return position >= 1 && position <= attributes.size();
    }
}
